@@ -3,12 +3,13 @@ from __future__ import annotations
 import itertools
 import logging
 import typing as t
+import uuid
 from collections import defaultdict
 from dataclasses import dataclass, field
 
 import pydantic
 from piccolo.columns import Column, Where
-from piccolo.columns.column_types import Array, Text, Varchar
+from piccolo.columns.column_types import Array, ForeignKey, Text, Varchar
 from piccolo.columns.operators import (
     Equal,
     GreaterEqualThan,
@@ -26,12 +27,19 @@ from starlette.requests import Request
 from starlette.responses import JSONResponse, Response
 from starlette.routing import Route, Router
 
+from piccolo_api.crud.hooks import (
+    Hook,
+    HookType,
+    execute_delete_hooks,
+    execute_patch_hooks,
+    execute_post_hooks,
+)
+
 from .exceptions import MalformedQuery
 from .serializers import Config, create_pydantic_model
 from .validators import Validators, apply_validators
 
 if t.TYPE_CHECKING:  # pragma: no cover
-    from piccolo.columns import Selectable
     from piccolo.query.methods.count import Count
     from piccolo.query.methods.objects import Objects
     from starlette.datastructures import QueryParams
@@ -52,6 +60,8 @@ OPERATOR_MAP = {
 
 MATCH_TYPES = ("contains", "exact", "starts", "ends")
 
+PK_TYPES = t.Union[str, uuid.UUID, int]
+
 
 class CustomJSONResponse(Response):
     media_type = "application/json"
@@ -76,6 +86,48 @@ class Params:
     include_readable: bool = False
     page: int = 1
     page_size: t.Optional[int] = None
+    visible_fields: str = field(default="")
+    range_header: bool = False
+    range_header_name: str = field(default="")
+
+
+def get_visible_fields_options(
+    table: t.Type[Table],
+    exclude_secrets: bool = False,
+    max_joins: int = 0,
+    prefix: str = "",
+) -> t.Tuple[str, ...]:
+    """
+    In the schema, we tell the user which fields are allowed with the
+    ``__visible_fields`` GET parameter. This function extracts the column
+    names, and names of related columns too.
+
+    :param prefix:
+        Used internally by this function - the user doesn't need to set this.
+
+    """
+    fields = []
+
+    for column in table._meta.columns:
+        if exclude_secrets and column._meta.secret:
+            continue
+
+        column_name = (
+            f"{prefix}.{column._meta.name}" if prefix else column._meta.name
+        )
+        fields.append(column_name)
+
+        if isinstance(column, ForeignKey) and max_joins > 0:
+            fields.extend(
+                get_visible_fields_options(
+                    table=column._foreign_key_meta.resolved_references,
+                    exclude_secrets=exclude_secrets,
+                    max_joins=max_joins - 1,
+                    prefix=column_name,
+                )
+            )
+
+    return tuple(fields)
 
 
 class PiccoloCRUD(Router):
@@ -92,30 +144,89 @@ class PiccoloCRUD(Router):
         allow_bulk_delete: bool = False,
         page_size: int = 15,
         exclude_secrets: bool = True,
-        validators: Validators = Validators(),
+        validators: t.Optional[Validators] = None,
+        schema_extra: t.Optional[t.Dict[str, t.Any]] = None,
+        max_joins: int = 0,
+        hooks: t.Optional[t.List[Hook]] = None,
     ) -> None:
         """
         :param table:
             The Piccolo ``Table`` to expose CRUD methods for.
         :param read_only:
-            If True, only the GET method is allowed.
+            If ``True``, only the GET method is allowed.
         :param allow_bulk_delete:
-            If True, allows a delete request to the root to delete all matching
-            records. It is dangerous, so is disabled by default.
+            If ``True``, allows a delete request to the root to delete all
+            matching records. It is dangerous, so is disabled by default.
         :param page_size:
             The number of results shown on each page by default.
         :param exclude_secrets:
-            Any values in Secret columns will be omitted from the response.
+            Any Piccolo columns with ``secret=True`` will be omitted from the
+            response.
         :param validators:
             Used to provide extra validation on certain endpoints - can be
             easier than subclassing.
-        """
+        :param schema_extra:
+            Additional information included in the Pydantic schema.
+        :param max_joins:
+            Determines whether a query can request data from related tables
+            using joins. For example ``/movie/?__visible_fields=name,director.name``,
+            which would return:
+
+            .. code-block:: javascript
+
+                {
+                    'rows': [
+                        {
+                            'name': 'Star Wars',
+                            'director': {
+                                'name': 'George Lucas'
+                            }
+                        }
+                    ]
+                }
+
+            This is a very powerful feature, but before enabling it, bear in
+            mind the following:
+
+            * If set too high, it could be used maliciously to craft slow
+              queries which contain lots of joins, which could slow down your
+              site.
+            * Don't enable it if sensitive data is contained in related
+              tables, as this feature can be used to retrieve that data.
+
+            It's best used when the data in related tables is not of a
+            sensitive nature and the client is highly trusted. Consider using
+            it with ``exclude_secrets=True``.
+
+            To see which fields can be filtered in this way, you can check
+            the ``visible_fields_options`` value returned by the ``/schema``
+            endpoint.
+
+        """  # noqa: E501
         self.table = table
         self.page_size = page_size
         self.read_only = read_only
         self.allow_bulk_delete = allow_bulk_delete
         self.exclude_secrets = exclude_secrets
         self.validators = validators
+        self.max_joins = max_joins
+        if hooks:
+            self._hook_map = {
+                group[0]: [hook for hook in group[1]]
+                for group in itertools.groupby(hooks, lambda x: x.hook_type)
+            }
+        else:
+            self._hook_map = None  # type: ignore
+
+        schema_extra = schema_extra if isinstance(schema_extra, dict) else {}
+        self.visible_fields_options = get_visible_fields_options(
+            table=table, exclude_secrets=exclude_secrets, max_joins=max_joins
+        )
+        schema_extra["visible_fields_options"] = self.visible_fields_options
+        schema_extra[
+            "primary_key_name"
+        ] = self.table._meta.primary_key._meta.name
+        self.schema_extra = schema_extra
 
         root_methods = ["GET"]
         if not read_only:
@@ -125,13 +236,6 @@ class PiccoloCRUD(Router):
 
         routes: t.List[BaseRoute] = [
             Route(path="/", endpoint=self.root, methods=root_methods),
-            Route(
-                path="/{row_id:int}/",
-                endpoint=self.detail,
-                methods=["GET"]
-                if read_only
-                else ["GET", "PUT", "DELETE", "PATCH"],
-            ),
             Route(path="/schema/", endpoint=self.get_schema, methods=["GET"]),
             Route(path="/ids/", endpoint=self.get_ids, methods=["GET"]),
             Route(path="/count/", endpoint=self.get_count, methods=["GET"]),
@@ -146,6 +250,13 @@ class PiccoloCRUD(Router):
                 endpoint=self.update_password,
                 methods=["PUT"],
             ),
+            Route(
+                path="/{row_id:str}/",
+                endpoint=self.detail,
+                methods=["GET"]
+                if read_only
+                else ["GET", "PUT", "DELETE", "PATCH"],
+            ),
         ]
 
         super().__init__(routes=routes)
@@ -158,17 +269,25 @@ class PiccoloCRUD(Router):
         Useful for serialising inbound data from POST and PUT requests.
         """
         return create_pydantic_model(
-            self.table, model_name=f"{self.table.__name__}In"
+            self.table,
+            model_name=f"{self.table.__name__}In",
+            exclude_columns=(self.table._meta.primary_key,),
+            **self.schema_extra,
         )
 
     def _pydantic_model_output(
-        self, include_readable: bool = False
+        self,
+        include_readable: bool = False,
+        include_columns: t.Tuple[Column, ...] = (),
+        nested: t.Union[bool, t.Tuple[Column, ...]] = False,
     ) -> t.Type[pydantic.BaseModel]:
         return create_pydantic_model(
             self.table,
             include_default_columns=True,
             include_readable=include_readable,
+            include_columns=include_columns,
             model_name=f"{self.table.__name__}Output",
+            nested=nested,
         )
 
     @property
@@ -192,15 +311,22 @@ class PiccoloCRUD(Router):
             model_name=f"{self.table.__name__}Optional",
         )
 
-    def pydantic_model_plural(self, include_readable=False):
+    def pydantic_model_plural(
+        self,
+        include_readable=False,
+        include_columns: t.Tuple[Column, ...] = (),
+        nested: t.Union[bool, t.Tuple[Column, ...]] = False,
+    ):
         """
         This is for when we want to serialise many copies of the model.
         """
-        base_model = create_pydantic_model(
+        base_model: t.Any = create_pydantic_model(
             self.table,
             include_default_columns=True,
             include_readable=include_readable,
+            include_columns=include_columns,
             model_name=f"{self.table.__name__}Item",
+            nested=nested,
         )
         return pydantic.create_model(
             str(self.table.__name__) + "Plural",
@@ -234,7 +360,7 @@ class PiccoloCRUD(Router):
 
         An optional 'search' GET parameter can be used to filter the results
         returned. Also, an optional 'limit' paramter can be used to specify
-        how many results should be returned.
+        how many results should be returned, and 'offset' for basic pagination.
 
         """
         readable = self.table.get_readable()
@@ -253,6 +379,17 @@ class PiccoloCRUD(Router):
         else:
             limit = "ALL"
 
+        offset = request.query_params.get("offset")
+        if offset is not None:
+            try:
+                offset = int(offset)
+            except ValueError:
+                return Response(
+                    "The offset must be an integer", status_code=400
+                )
+        else:
+            offset = 0
+
         search_term = request.query_params.get("search")
         if search_term is not None:
             # Readable doesn't currently have a 'like' method, so we do it
@@ -264,7 +401,7 @@ class PiccoloCRUD(Router):
                         (
                             f"SELECT * FROM ({query.__str__()}) as subquery "
                             "WHERE subquery.readable ILIKE {} "
-                            f"LIMIT {limit}"
+                            f"LIMIT {limit} OFFSET {offset}"
                         ),
                         f"%{search_term}%",
                     ),
@@ -277,16 +414,25 @@ class PiccoloCRUD(Router):
                     "WHERE UPPER(subquery.readable) LIKE {}"
                 )
                 if isinstance(limit, int):
-                    sql += f" LIMIT {limit}"
+                    sql += f" LIMIT {limit} OFFSET {offset}"
                 query = t.cast(
                     Select, self.table.raw(sql, f"%{search_term.upper()}%")
                 )
         else:
             if limit != "ALL":
-                query = query.limit(limit)
+                query = query.limit(limit).offset(offset)
 
         values = await query.run()
-        return JSONResponse({i["id"]: i["readable"] for i in values})
+
+        primary_key = self.table._meta.primary_key
+        if primary_key.value_type not in (int, str):
+            return JSONResponse(
+                {str(i[primary_key._meta.name]): i["readable"] for i in values}
+            )
+        else:
+            return JSONResponse(
+                {i[primary_key._meta.name]: i["readable"] for i in values}
+            )
 
     ###########################################################################
 
@@ -346,10 +492,16 @@ class PiccoloCRUD(Router):
             for i in itertools.groupby(params.multi_items(), lambda x: x[0])
         }
 
+        array_columns = [
+            i._meta.name
+            for i in self.table._meta.columns
+            if i.value_type == list
+        ]
+
         output = {}
 
         for key, value in params_map.items():
-            if key.endswith("[]"):
+            if key.endswith("[]") or key.rstrip("[]") in array_columns:
                 # Is either an array, or multiple values have been passed in
                 # for another field.
                 key = key.rstrip("[]")
@@ -396,6 +548,16 @@ class PiccoloCRUD(Router):
 
         And can specify which page: {'__page': 2}.
 
+        You can specify which fields want to display in rows:
+        {'__visible_fields': 'id,name'}.
+
+        You can activate the "Content-Range" response header:
+        {'__range_header': True}
+
+        If the "Content-Range" response header is enabled,
+        you can configure the "plural name" used in the header:
+        {'__range_header_name': 'movies'}
+
         This method splits the params into their different types.
         """
         response = Params()
@@ -439,10 +601,23 @@ class PiccoloCRUD(Router):
                     response.page_size = page_size
                 continue
 
+            if key == "__visible_fields":
+                response.visible_fields = value
+                continue
+
             if key == "__readable" and value in ("true", "True", "1"):
                 response.include_readable = True
                 continue
 
+            if key == "__range_header":
+                if value in ("true", "True", "1"):
+                    response.range_header = True
+                continue
+
+            if key == "__range_header_name":
+                response.range_header_name = value
+                continue
+
             response.fields[key] = value
 
         return response
@@ -458,7 +633,6 @@ class PiccoloCRUD(Router):
         Objects etc.
         """
         fields = params.fields
-
         if fields:
             model_dict = self.pydantic_model_optional(**fields).dict()
             for field_name in fields.keys():
@@ -506,21 +680,44 @@ class PiccoloCRUD(Router):
 
         split_params = self._split_params(params)
 
-        include_readable = split_params.include_readable
-        if include_readable:
-            readable_columns = [
-                self.table._get_related_readable(i)
-                for i in self.table._meta.foreign_key_columns
-            ]
-            columns: t.List[Selectable] = [
-                *self.table._meta.columns,
-                *readable_columns,
-            ]
-            query = self.table.select(
-                *columns, exclude_secrets=self.exclude_secrets
+        # Visible fields
+        visible_fields = split_params.visible_fields
+        nested: t.Union[bool, t.Tuple[Column, ...]]
+        if visible_fields:
+            try:
+                visible_columns = self._parse_visible_fields(visible_fields)
+            except ValueError as exception:
+                return Response(str(exception), status_code=400)
+
+            nested = tuple(
+                i for i in visible_columns if len(i._meta.call_chain) > 0
             )
         else:
-            query = self.table.select(exclude_secrets=self.exclude_secrets)
+            visible_columns = self.table._meta.columns
+            nested = False
+
+        # Readable
+        include_readable = split_params.include_readable
+        readable_columns = (
+            [
+                self.table._get_related_readable(i)
+                for i in visible_columns
+                if isinstance(i, ForeignKey)
+            ]
+            if include_readable
+            else []
+        )
+
+        # Build select query, and exclude secrets
+        query = self.table.select(
+            *visible_columns,
+            *readable_columns,
+            exclude_secrets=self.exclude_secrets,
+        )
+
+        # Make it nested if required
+        if nested:
+            query = query.output(nested=True)
 
         # Apply filters
         try:
@@ -548,17 +745,38 @@ class PiccoloCRUD(Router):
             )
         query = query.limit(page_size)
         page = split_params.page
+        offset = 0
         if page > 1:
             offset = page_size * (page - 1)
             query = query.offset(offset).limit(page_size)
 
         rows = await query.run()
+        headers = {}
+        if split_params.range_header is True:
+            plural_name = (
+                split_params.range_header_name or self.table._meta.tablename
+            )
+
+            row_length = len(rows)
+            if row_length == 0:
+                curr_page_len = 0
+            else:
+                curr_page_len = row_length - 1
+            curr_page_len = curr_page_len + offset
+            count = await self.table.count().run()
+            curr_page_string = f"{offset}-{curr_page_len}"
+            headers[
+                "Content-Range"
+            ] = f"{plural_name} {curr_page_string}/{count}"
+
         # We need to serialise it ourselves, in case there are datetime
         # fields.
-        json = self.pydantic_model_plural(include_readable=include_readable)(
-            rows=rows
-        ).json()
-        return CustomJSONResponse(json)
+        json = self.pydantic_model_plural(
+            include_readable=include_readable,
+            include_columns=tuple(visible_columns),
+            nested=nested,
+        )(rows=rows).json()
+        return CustomJSONResponse(json, headers=headers)
 
     ###########################################################################
 
@@ -566,7 +784,11 @@ class PiccoloCRUD(Router):
         cleaned_data: t.Dict[str, t.Any] = {}
 
         for key, value in data.items():
-            value = None if value == "null" else value
+            value = (
+                None
+                if (isinstance(value, str) and value.lower() == "null")
+                else value
+            )
             cleaned_data[key] = value
 
         return cleaned_data
@@ -586,6 +808,10 @@ class PiccoloCRUD(Router):
 
         try:
             row = self.table(**model.dict())
+            if self._hook_map:
+                row = await execute_post_hooks(
+                    hooks=self._hook_map, hook_type=HookType.pre_save, row=row
+                )
             response = await row.save().run()
             json = dump_json(response)
             # Returns the id of the inserted row.
@@ -643,6 +869,11 @@ class PiccoloCRUD(Router):
         if row_id is None:
             return Response("Missing ID parameter.", status_code=404)
 
+        try:
+            row_id = self.table._meta.primary_key.value_type(row_id)
+        except ValueError:
+            return Response("The ID is invalid", status_code=400)
+
         if (
             not await self.table.exists()
             .where(self.table._meta.primary_key == row_id)
@@ -668,43 +899,103 @@ class PiccoloCRUD(Router):
         else:
             return Response(status_code=405)
 
+    def _parse_visible_fields(self, visible_fields: str) -> t.List[Column]:
+        """
+        Parse the ``visible_fields`` string, and return a list of columns.
+
+        :param visible_fields:
+            A comma separated list of column names, for example ``'id,name'``.
+            The presence of a full stop in the name indicates a join, for
+            example ``'director.name'``.
+        :raises ValueError:
+            If the max join depth is exceeded, or the column name isn't
+            recognised.
+
+        """
+        column_names: t.List[str] = visible_fields.split(",")
+        visible_columns: t.List[Column] = []
+
+        for column_name in column_names:
+            try:
+                column = self.table._meta.get_column_by_name(column_name)
+            except ValueError as exception:
+                raise ValueError(
+                    f"{exception} - the column options are "
+                    f"{self.visible_fields_options}."
+                )
+
+            if len(column._meta.call_chain) > self.max_joins:
+                raise ValueError("Max join depth exceeded")
+            else:
+                visible_columns.append(column)
+
+        return visible_columns
+
     @apply_validators
-    async def get_single(self, request: Request, row_id: int) -> Response:
+    async def get_single(self, request: Request, row_id: PK_TYPES) -> Response:
         """
         Returns a single row.
         """
         params = dict(request.query_params)
         split_params: Params = self._split_params(params)
-        try:
-            columns: t.Sequence[Selectable] = self.table._meta.columns
-            if split_params.include_readable:
-                readable_columns = [
-                    self.table._get_related_readable(i)
-                    for i in self.table._meta.foreign_key_columns
-                ]
-                columns = [*columns, *readable_columns]
 
-            row = (
-                await self.table.select(
-                    *columns, exclude_secrets=self.exclude_secrets
-                )
-                .where(self.table._meta.primary_key == row_id)
-                .first()
-                .run()
+        # Visible fields
+        nested: t.Union[bool, t.Tuple[Column, ...]]
+        visible_fields = split_params.visible_fields
+        if visible_fields:
+            try:
+                visible_columns = self._parse_visible_fields(visible_fields)
+            except ValueError as exception:
+                return Response(str(exception), status_code=400)
+
+            nested = tuple(
+                i for i in visible_columns if len(i._meta.call_chain) > 0
             )
-        except ValueError:
+        else:
+            visible_columns = self.table._meta.columns
+            nested = False
+
+        # Readable
+        readable_columns = (
+            [
+                self.table._get_related_readable(i)
+                for i in self.table._meta.foreign_key_columns
+            ]
+            if split_params.include_readable
+            else []
+        )
+
+        query = (
+            self.table.select(
+                *visible_columns,
+                *readable_columns,
+                exclude_secrets=self.exclude_secrets,
+            )
+            .where(self.table._meta.primary_key == row_id)
+            .first()
+        )
+
+        if nested:
+            query = query.output(nested=True)
+
+        row = await query.run()
+
+        if not row:
             return Response(
                 "Unable to find a resource with that ID.", status_code=404
             )
+
         return CustomJSONResponse(
             self._pydantic_model_output(
-                include_readable=split_params.include_readable
+                include_readable=split_params.include_readable,
+                include_columns=tuple(visible_columns),
+                nested=nested,
             )(**row).json()
         )
 
     @apply_validators
     async def put_single(
-        self, request: Request, row_id: int, data: t.Dict[str, t.Any]
+        self, request: Request, row_id: PK_TYPES, data: t.Dict[str, t.Any]
     ) -> Response:
         """
         Replaces an existing row. We don't allow new resources to be created.
@@ -723,6 +1014,7 @@ class PiccoloCRUD(Router):
         }
 
         try:
+
             await cls.update(values).where(
                 cls._meta.primary_key == row_id
             ).run()
@@ -732,7 +1024,7 @@ class PiccoloCRUD(Router):
 
     @apply_validators
     async def patch_single(
-        self, request: Request, row_id: int, data: t.Dict[str, t.Any]
+        self, request: Request, row_id: PK_TYPES, data: t.Dict[str, t.Any]
     ) -> Response:
         """
         Patch a single row.
@@ -756,6 +1048,14 @@ class PiccoloCRUD(Router):
                 f"Unrecognised keys - {unrecognised_keys}.", status_code=400
             )
 
+        if self._hook_map:
+            values = await execute_patch_hooks(
+                hooks=self._hook_map,
+                hook_type=HookType.pre_patch,
+                row_id=row_id,
+                values=values,
+            )
+
         try:
             await cls.update(values).where(
                 cls._meta.primary_key == row_id
@@ -766,20 +1066,30 @@ class PiccoloCRUD(Router):
                 .first()
                 .run()
             )
-            return JSONResponse(self.pydantic_model(**new_row).json())
+            return CustomJSONResponse(self.pydantic_model(**new_row).json())
         except ValueError:
             return Response("Unable to save the resource.", status_code=500)
 
     @apply_validators
-    async def delete_single(self, request: Request, row_id: int) -> Response:
+    async def delete_single(
+        self, request: Request, row_id: PK_TYPES
+    ) -> Response:
         """
         Deletes a single row.
         """
+
+        if self._hook_map:
+            await execute_delete_hooks(
+                hooks=self._hook_map,
+                hook_type=HookType.pre_delete,
+                row_id=row_id,
+            )
+
         try:
             await self.table.delete().where(
                 self.table._meta.primary_key == row_id
             ).run()
-            return Response("Deleted the resource.", status_code=204)
+            return Response(status_code=204)
         except ValueError:
             return Response("Unable to delete the resource.", status_code=500)
 
