@@ -6,12 +6,11 @@ from __future__ import annotations
 
 import typing as t
 from collections import defaultdict
-from dataclasses import dataclass, field
 from decimal import Decimal
 from enum import Enum
 from inspect import Parameter, Signature
 
-from fastapi import FastAPI, Request
+from fastapi import APIRouter, FastAPI, Request
 from fastapi.params import Query
 from pydantic import BaseModel as PydanticBaseModel
 from pydantic.main import BaseModel
@@ -26,20 +25,30 @@ class HTTPMethod(str, Enum):
     delete = "DELETE"
 
 
-@dataclass
 class FastAPIKwargs:
     """
     Allows kwargs to be passed into ``FastAPIApp.add_api_route``.
     """
 
-    all_routes: t.Dict[str, t.Any] = field(default_factory=dict)
-    get: t.Dict[str, t.Any] = field(default_factory=dict)
-    delete: t.Dict[str, t.Any] = field(default_factory=dict)
-    post: t.Dict[str, t.Any] = field(default_factory=dict)
-    put: t.Dict[str, t.Any] = field(default_factory=dict)
-    patch: t.Dict[str, t.Any] = field(default_factory=dict)
-    get_single: t.Dict[str, t.Any] = field(default_factory=dict)
-    delete_single: t.Dict[str, t.Any] = field(default_factory=dict)
+    def __init__(
+        self,
+        all_routes: t.Dict[str, t.Any] = {},
+        get: t.Dict[str, t.Any] = {},
+        delete: t.Dict[str, t.Any] = {},
+        post: t.Dict[str, t.Any] = {},
+        put: t.Dict[str, t.Any] = {},
+        patch: t.Dict[str, t.Any] = {},
+        get_single: t.Dict[str, t.Any] = {},
+        delete_single: t.Dict[str, t.Any] = {},
+    ):
+        self.all_routes = all_routes
+        self.get = get
+        self.delete = delete
+        self.post = post
+        self.put = put
+        self.patch = patch
+        self.get_single = get_single
+        self.delete_single = delete_single
 
     def get_kwargs(self, endpoint_name: str) -> t.Dict[str, t.Any]:
         """
@@ -76,7 +85,7 @@ class FastAPIWrapper:
     boilerplate code.
 
     :param root_url:
-        The URL to mount the endpoint at - e.g. /movies/.
+        The URL to mount the endpoint at - e.g. ``'/movies/'``.
     :param fastapi_app:
         The ``FastAPI`` instance you want to attach the endpoints to.
     :param piccolo_crud:
@@ -84,17 +93,19 @@ class FastAPIWrapper:
         the arguments passed into ``PiccoloCRUD``, for example ``ready_only``
         and ``allow_bulk_delete``.
     :param fastapi_kwargs:
-        Specifies the extra kwargs to pass to FastAPI's `add_api_route`.
+        Specifies the extra kwargs to pass to FastAPI's ``add_api_route``.
 
     """
 
     def __init__(
         self,
         root_url: str,
-        fastapi_app: FastAPI,
+        fastapi_app: t.Union[FastAPI, APIRouter],
         piccolo_crud: PiccoloCRUD,
-        fastapi_kwargs: FastAPIKwargs = FastAPIKwargs(),
+        fastapi_kwargs: t.Optional[FastAPIKwargs] = None,
     ):
+        fastapi_kwargs = fastapi_kwargs or FastAPIKwargs()
+
         self.root_url = root_url
         self.fastapi_app = fastapi_app
         self.piccolo_crud = piccolo_crud
@@ -283,14 +294,14 @@ class FastAPIWrapper:
         #######################################################################
         # Detail - GET
 
-        async def get_single(row_id: int, request: Request):
+        async def get_single(row_id: str, request: Request):
             """
             Retrieve a single row from the table.
             """
             return await piccolo_crud.detail(request=request)
 
         fastapi_app.add_api_route(
-            path=self.join_urls(root_url, "/{row_id:int}/"),
+            path=self.join_urls(root_url, "/{row_id:str}/"),
             endpoint=get_single,
             response_model=self.ModelOut,
             methods=["GET"],
@@ -302,14 +313,14 @@ class FastAPIWrapper:
 
         if not piccolo_crud.read_only:
 
-            async def delete_single(row_id: int, request: Request):
+            async def delete_single(row_id: str, request: Request):
                 """
                 Delete a single row from the table.
                 """
                 return await piccolo_crud.detail(request=request)
 
             fastapi_app.add_api_route(
-                path=self.join_urls(root_url, "/{row_id:int}/"),
+                path=self.join_urls(root_url, "/{row_id:str}/"),
                 endpoint=delete_single,
                 response_model=None,
                 methods=["DELETE"],
@@ -321,7 +332,7 @@ class FastAPIWrapper:
 
         if not piccolo_crud.read_only:
 
-            async def put(row_id: int, request: Request, model):
+            async def put(row_id: str, request: Request, model):
                 """
                 Insert or update a single row.
                 """
@@ -332,7 +343,7 @@ class FastAPIWrapper:
             ] = f"ANNOTATIONS['{self.alias}']['ModelIn']"
 
             fastapi_app.add_api_route(
-                path=self.join_urls(root_url, "/{row_id:int}/"),
+                path=self.join_urls(root_url, "/{row_id:str}/"),
                 endpoint=put,
                 response_model=self.ModelOut,
                 methods=["PUT"],
@@ -344,7 +355,7 @@ class FastAPIWrapper:
 
         if not piccolo_crud.read_only:
 
-            async def patch(row_id: int, request: Request, model):
+            async def patch(row_id: str, request: Request, model):
                 """
                 Update a single row.
                 """
@@ -355,7 +366,7 @@ class FastAPIWrapper:
             ] = f"ANNOTATIONS['{self.alias}']['ModelOptional']"
 
             fastapi_app.add_api_route(
-                path=self.join_urls(root_url, "/{row_id:int}/"),
+                path=self.join_urls(root_url, "/{row_id:str}/"),
                 endpoint=patch,
                 response_model=self.ModelOut,
                 methods=["PATCH"],
@@ -398,7 +409,7 @@ class FastAPIWrapper:
         ]
 
         for field_name, _field in model.__fields__.items():
-            type_ = _field.type_
+            type_ = _field.outer_type_
             parameters.append(
                 Parameter(
                     name=field_name,
@@ -489,9 +500,56 @@ class FastAPIWrapper:
                                 ),
                             ),
                         ),
+                        Parameter(
+                            name="__visible_fields",
+                            kind=Parameter.POSITIONAL_OR_KEYWORD,
+                            annotation=str,
+                            default=Query(
+                                default=None,
+                                description=(
+                                    "The fields to return. It's a comma "
+                                    "separated list - for example "
+                                    "'name,address'. By default all fields "
+                                    "are returned."
+                                ),
+                            ),
+                        ),
                     ]
                 )
 
+            parameters.extend(
+                [
+                    Parameter(
+                        name="__range_header",
+                        kind=Parameter.POSITIONAL_OR_KEYWORD,
+                        annotation=bool,
+                        default=Query(
+                            default=False,
+                            description=(
+                                "Set to 'true' to add the "
+                                "Content-Range response header"
+                            ),
+                        ),
+                    )
+                ]
+            )
+            parameters.extend(
+                [
+                    Parameter(
+                        name="__range_header_name",
+                        kind=Parameter.POSITIONAL_OR_KEYWORD,
+                        annotation=str,
+                        default=Query(
+                            default=None,
+                            description=(
+                                "Specify the object name in the Content-Range "
+                                "response header (defaults to the table name)."
+                            ),
+                        ),
+                    )
+                ]
+            )
+
         endpoint.__signature__ = Signature(  # type: ignore
             parameters=parameters
         )
