@@ -1,20 +1,21 @@
 from __future__ import annotations
 
-from abc import abstractmethod, ABCMeta
 import typing as t
+from abc import ABCMeta, abstractmethod
 
 from piccolo.apps.user.tables import BaseUser as BaseUserTable
-from piccolo_api.token_auth.tables import TokenAuth
-from piccolo_api.shared.auth import User
 from starlette.authentication import (
-    AuthenticationBackend,
     AuthCredentials,
+    AuthenticationBackend,
     AuthenticationError,
-    SimpleUser,
     BaseUser,
+    SimpleUser,
 )
 from starlette.requests import HTTPConnection
 
+from piccolo_api.shared.auth import User
+from piccolo_api.token_auth.tables import TokenAuth
+
 
 class TokenAuthProvider(metaclass=ABCMeta):
     """
@@ -65,7 +66,7 @@ class PiccoloTokenAuthProvider(TokenAuthProvider):
 
         user = (
             await self.auth_table.select(self.auth_table.username)
-            .where(self.auth_table.id == user_id)
+            .where(self.auth_table._meta.primary_key == user_id)
             .first()
             .run()
         )
